/*****************************************************************************
 * Copyright (C) 1997-2007, Mark Hummel
 * This file is part of Vrq.
 *
 * Vrq is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * Vrq is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, 
 * Boston, MA  02110-1301  USA
 *****************************************************************************
 */
/******************************************************************************
 *
 *
 *	   cdecl.hpp
 *		- abstract class for declaration 
 *
 *
 ******************************************************************************
 */

#ifndef CDECL_HPP
#define CDECL_HPP

#include <stdio.h>
#include <vector>
#include <list>
#include "glue.h"
#include "csymbol.h"

class CNode;

/**
 * Structure to hold file coordinates
 */
struct Coord_t {
	int	    lineno;	///< file line number
	const char* filename;	///< file name
};


/**
 * Expression node type
 */
enum NodeType_t {
	eU,	///< undefined
	eR,	///< real - have width 0
	eS,	///< signed bit vector, includes integer
	eE,	///< event - have width 0
	eB	///< unsigned bit vector
};


/**
 * Declaration types
 */
enum Decl_t {
	eREG = 0,	///< register declaration
	eREAL,		///< real declaration
	eREALTIME,	///< realtime declaration
	eTIME,		///< time declaration
	eINT,		///< integer declaration
	eWIRE,		///< wire declaration
	eTRI,		///< tri declaration
	eWAND,		///< wand declaration
	eTRIAND,	///< triand declaration
	eWOR,		///< wor declaration
	eTRIOR,		///< trior declaration
	eTRI1,		///< tri1 declaration
	eTRI0,		///< tri0 declaration
	eTRIREG, 	///< trireg declaration
	eSUPPLY0,	///< supply0 declaration
	eSUPPLY1,	///< supply1 declaration
	ePARAM,		///< parameter declaration
	eLOCALPARAM,	///< localparam declaration
	eBLOCK,		///< named block declaration
	ePARALLELBLOCK, ///< fork/join block declaration	
	eGENVARBLOCK,	///< generate block declaration
	eMODULE,	///< module declaration
	eMACRO,		///< macro module declaration
	eINSTANCE,	///< instance declaration
	eINPUT,		///< input declaration
	eOUTPUT,	///< output declaration
	eINOUT,		///< inout declaration
	ePORT,		///< port declaration
	eFREF,		///< forward reference declaration (used by parser only)
	eFUNCTION,	///< function/task declaration
	eSPECIFYBLOCK,	///< specify block declaration
	eGENVAR,	///< genvar declaration
	eDEVENT, 	///< event declaration
	// NOTE this are class types only
	eNET,		///< declaration superclass for all net types
	ePORTDIR,	///< declaration superclass for all port dirs
	eATTR,		///< attribute declaration
	eUDP,		///< udp declaration
	eNONE		///< unspecified declaration used for wildcard
};

#if defined CDECL_CPP
/**
 * Array to convert Decl_t to character string
 */
const char* declName[] = {
	"reg",
	"real",
	"realtime",
	"time",
	"integer",
	"wire",
	"tri",
	"wand",
	"triand",
	"wor",
	"trior",
	"tri1",
	"tri0",
	"trireg", 
	"supply0",
	"supply1",
	"param",
	"localparam",
	"block",
	"parallel block",
	"genvar block",
	"module",
	"macro",
	"instance",
	"input",
	"output",
	"inout",
	"port",
	"fref",
	"function",
	"specify block",
	"genvar",
	"event",
	"net",
	"portdir",
	"attr",
	"upd",
	"none"
};
#else
extern char* declName[];
#endif



/**
 * Base class for describing declaration objects
 */
class CDecl : public CObject
{
public:
	enum Flag {
	    eFLAG_NONE = 0,
	    eFLAG_PRAGMA = 1,
	    eFLAG_ARRAY = 2,
	    eFLAG_VECTOR = 4
	};
	static Flag Or( Flag f1, Flag f2 ) 
		{ return static_cast<Flag>(((unsigned)f1 | (unsigned)f2)); }
	static Flag Or( Flag f1, Flag f2, Flag f3 ) 
		{ return static_cast<Flag>(((unsigned)f1 | (unsigned)f2) | (unsigned)f3); }
private:
	Coord_t	   loc;		  ///< file coordinates of declaration
	CSymbol*   symbol;	  ///< symbol for declaration
	Decl_t	    type;	  ///< declaration type
	int	   declStatement; ///< TRUE if decl statement was created 
				  ///< in containing block
	CNode*	   attributes;	  ///< declaration attributes
	CNode*	   pragmas;	  ///< declaration pragmas
	vector<CNode*> indexes;	  ///< array limits
	int        signed_;	  ///< TRUE if signed.
	CNode*     range;	  ///< upper/lower range of vector.
	int numberOfDimensions;	  ///< array dimensions
	Flag	   flags;	  ///< declaration support flags.
public:
	/**
 	 * Determine if width of declaration is constant, ie dependent
 	 * upon only constants and parameters.
 	 * \return non-zero of width of declaration is constant.
 	 */
	virtual int  IsWidthConstant( void );
	/**
 	 * Determine if width of declaration is volatile, ie depend
 	 * upon parameters or variables.
 	 * \return non-zero of width of declaration is volatile.
 	 */
	virtual int  IsWidthVolatile( void );
	/**
 	 * Determine if width of declaration can be evaluated.
 	 * \return non-zero of width of declaration can be evaluated.
 	 */
	virtual int  IsWidthEvaluateable( void );
	/**
 	 * Evaluate width of declaration.
 	 * \return width of declaration.
 	 */
	virtual INT32 GetWidth( void );
	/**
 	 * Get expression representing width of declaration.
 	 * \return expression representing width of declaration.
	 */
	virtual CNode* GetWidthExp( void );
	/**
 	 * Evaluate current decl width direction. Expression must be constant.  
	 * \return 1 - big endian, -1 - little endian, 0 - cannot be determined
	 */
	virtual int   WidthDirection( void );
	/**
 	 * Evaluate the msb of the declaration.
 	 * \return msb of declaration, 0 if scalar.
 	 */
	virtual INT32 GetMsbInt( void );
        /**
         * Get expression for declaration's msb
         * \return msb expression, NULL if scalar.
         */
        virtual CNode* GetMsb();
        /**
         * Get expression for declaration's range (msb/lsb)
         * \return range expression, NULL if scalar.
         */
        virtual CNode* GetRange() { return range; }
	/**
 	 * Set declaration's range (msb/lsb).
 	 * \param n range expression or NULL
 	 */
	virtual void SetRange( CNode* n );
	/**
 	 * Evaluate the lsb of the declaration.
 	 * \return lsb of declaration, 0 if scalar.
 	 */
	virtual INT32 GetLsbInt( void );
        /**
         * Get expression for declaration's lsb
         * \return lsb expression, NULL if scalar.
         */
        virtual CNode* GetLsb();
	/**
 	 * Get number of dimension of declaration.
 	 * \return number of declarations, 0 if non-array.
 	 */
	virtual INT32 GetNumberOfDimensions( void ) { 
	    return numberOfDimensions;
	}
	/**
 	 * Get expression tree for upper limit of array dimension.
 	 * \param dim dimension.
 	 * \return expression tree for upper limit.
	 */
	virtual CNode* GetMsi( INT32 dim );
	/**
 	 * Get expression tree for lower limit of array dimension.
 	 * \param dim dimension.
 	 * \return expression tree for lower limit.
	 */
	virtual CNode* GetLsi( INT32 dim );
	/**
 	 * Set number of dimensions for declaration.
 	 * \param dim number of dimensions.
 	 */
	virtual void SetNumberOfDimensions( INT32 dim ) { 
	    MASSERT( flags & eFLAG_ARRAY );
	    numberOfDimensions = dim;
	    indexes.reserve( dim );
	}
	/**
 	 * Get expression for limit of array for dimension.
 	 * \param dim dimension to get.
 	 * \return range expression for limit.
 	 */
	virtual CNode*  GetIndex( INT32 dim ) { return indexes[dim]; }
	/**
 	 * Set expression for limits of array for declaration.
 	 * \param dim dimension to set.
 	 * \param v range expression for limit.
 	 */
	virtual void  SetIndex( INT32 dim, CNode* v ) 
		{ MASSERT(flags & eFLAG_ARRAY); indexes[dim] = v; }
	/**
 	 * Set declartion's signed property.
 	 * \param v value of signed property.
 	 */
	virtual void SetSigned( int v ) 
			{ MASSERT( flags & eFLAG_VECTOR ); signed_ = v; }
	/**
 	 * Get declartion's signed property.
 	 * \return value of signed property.
 	 */
	virtual int  GetSigned() { return signed_; }
	/**
 	 * Set declartion's vectored property.
 	 * \param v value of vectored property.
 	 */
	virtual void SetVectored( int v ) { MASSERT( FALSE ); }
	/**
 	 * Get declartion's vectored property.
 	 * \return value of vectored property.
 	 */
	virtual int  GetVectored() { MASSERT( FALSE ); return 0; }
	/**
 	 * Set declartion's scalared property.
 	 * \param v value of scalared property.
 	 */
	virtual void SetScalared( int v ) { MASSERT( FALSE ); }
	/**
 	 * Get declartion's scalared property.
 	 * \return value of scalared property.
 	 */
	virtual int  GetScalared() { MASSERT( FALSE ); return 0; }
	/**
 	 * Set declarations's attributes.
 	 * \param attr attribute list.
 	 */
	void	SetAttributes( CNode* attr ) { attributes = attr; }
	/**
 	 * Get declaration's attributes.
 	 * \return attribute list.
 	 */
	CNode*  GetAttributes() { return attributes; }
	/**
 	 * Determine if declaration has the given attribute.
 	 * \param name name of attribute to search for.
 	 * \param n don't supply this argument.
 	 * \param init don't supply this argument.
 	 */
        int     HasAttribute( char* name, CNode* n=NULL, int init = 1 );
	/**
 	 * Get node type of decl.
 	 * \return node type of decl.
 	 */
	virtual NodeType_t GetNodeType( void ) { return eU; }
        /**
         * Get class of declaration.
         * \return declaration class for this object.
         */
        Decl_t	GetClass( void )
        {
        	switch( type ) {
        	case ePARAM:
        	case eLOCALPARAM:
        	       	return ePARAM;
        	case eREAL:
        	case eREALTIME:
			return eREAL;
        	case eREG:
        	case eTIME:
        	case eINT:
        		return eREG;
        	case eWIRE:
        	case eTRI:
        	case eWAND:
        	case eTRIAND:
        	case eWOR:
        	case eTRIOR:
        	case eTRI1:
        	case eTRI0:
        	case eTRIREG: 
        	case eSUPPLY0:
        	case eSUPPLY1:
        		return eNET;
        	case eINPUT:
        	case eOUTPUT:
        	case eINOUT:
        		return ePORTDIR;
        	default:
        		return type;
        	}
        }

        /**
         * Get a list of members of the given declaration class/type.
         * \param type type/class of declaration.
         * \param result list to return members.
         */
        static void GetMembers( Decl_t type, list<Decl_t>& result )
        {
            switch( type ) {
            case ePARAM:
                result.push_back( ePARAM );
                result.push_back( eLOCALPARAM );
                break;
	    case eREAL:
                result.push_back( eREAL );
                result.push_back( eREALTIME );
            case eREG:
                result.push_back( eREG );
                result.push_back( eTIME );
                result.push_back( eINT );
                break;	
            case eNET:
                result.push_back( eWIRE );
                result.push_back( eTRI );
                result.push_back( eWAND );
                result.push_back( eTRIAND );
                result.push_back( eWOR );
                result.push_back( eTRIOR );
                result.push_back( eTRI1 );
                result.push_back( eTRI0 );
                result.push_back( eTRIREG );
                result.push_back( eSUPPLY0 );
                result.push_back( eSUPPLY1 );
                break;
            case ePORTDIR:
                result.push_back( eINPUT );
                result.push_back( eOUTPUT );
                result.push_back( eINOUT );
                break;
            default:
                result.push_back( type );
                break;
            }
        }

        /**
         * Set declaration statement created attribute
         * \todo This call will be deprecated when the Block:Add(*) is
         * removed.
         */
        void SetDeclStatementCreated( void ) { declStatement = TRUE; }
        
        /**
         * Get declaration statement create attibute.
         * \return non-zero if declaration statement has been created.
         */
        int DeclStatementCreated( void ) { return declStatement; }

        /**
         * Get declaration type
         * \return declaration type.
         */
        Decl_t GetType( void ) { return type; }
        
        /**
         * Set declaration coordinate.
         * \param aLoc set file coordinates for declaration
         */
        void SetCoord( Coord_t* aLoc ) { loc = *aLoc; }
        
        /**
         * Get file coordinates for declaration.
         * \return file coordinates.
         */
        Coord_t*	GetCoord( void ) {
        	return &loc;
        }
        
        /**
         * Dump declaration info to file descriptor.
         * \param f file descriptor.
         */
        virtual void Dump( FILE* f ) 
        	{ fprintf( f, "'%s' line %d", loc.filename , loc.lineno ); }

	/**
	* Dump declaration name, type and location to file descriptor.
	* \param f file descriptor.
	*/
	virtual void DumpDeclInfo( FILE* f )
	{
	    fprintf( f, "%s: %s, defined in '%s' line %d\n", 
		declName[GetType()], GetName(), loc.filename , loc.lineno );
	}

        /**
         * Shortcut to get declaration's name
         * \return char pointer to name.
         */
        const char* GetName( void ) { return symbol->GetName(); }
        /**
         * Set declaration's symbol
         * \param aSymbol new symbol for declaration.
         */
        void SetSymbol( CSymbol* aSymbol ) { symbol = aSymbol; }
        /**
         * Get declaration's symbol
         * \return declaration's symbol.
         */
        CSymbol* GetSymbol( void ) { return symbol; }
	/**
 	 * Set declaration pragmas
 	 * \param p pragma list.
 	 */
	void    SetPragmas( CNode* p ) 
		{ MASSERT( flags & eFLAG_PRAGMA );pragmas = p; }
	/**
 	 * Get declaration pragmas
 	 * \return declaration's pragma list.
 	 */
	CNode*  GetPragmas() { return pragmas; }
protected:
        /**
         * Create instance of declaration.
         * This should never be call directly.
         * \param aSymbol declaration for symbol.
         * \param aLoc file coordinates of declaration.
         * \param aType declaration type.
         * \param flags list of supported constructs.
         */
        CDecl( CSymbol* aSymbol, Coord_t* aLoc, 
	       Decl_t aType, Flag flags )  {
        	loc 	 = *aLoc;
        	symbol   = aSymbol;	
        	type 	 = aType;
	        numberOfDimensions = 0;
        	declStatement  = FALSE;
        	attributes = NULL;
		pragmas = NULL;
		range = NULL;
		signed_ = FALSE;
		this->flags = flags;
        }
	/**
 	 * Perform deep copy of given object to this one
 	 * This should never be call directly, only by subclasses.
 	 * \param o declaration to copy.
 	 */
	void Copy( const CDecl& o );
	/**
 	 * Copy constructor to do deep copy.
 	 * This should never be called directly.
 	 * \param o declaration to copy.
 	 */
	CDecl( const CDecl& o ) { Copy( o ); }

public:
	virtual void PreVisit1( int (*func)(CNode*,void*), void* data );
	virtual void PostVisit1( void (*func)(CNode*, void*), void* data );
	virtual void PostSubVisit1( CNode* (*func)(CNode*, void*), void* data );

};

#endif // CDECL_HPP

