/*****************************************************************************
 * Copyright (C) 1997-2007, Mark Hummel
 * This file is part of Vrq.
 *
 * Vrq is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * Vrq is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, 
 * Boston, MA  02110-1301  USA
 *****************************************************************************
 */
/******************************************************************************
 *
 *
 *	   cparam.hpp
 *		- class definition of parameter 
 *		  definition nodes
 *
 ******************************************************************************
 */

#ifndef CPARAM_HPP
#define CPARAM_HPP

#include <stdio.h>
#include "glue.h"
#include "cnode.h"
#include "cdecl.h"


class CNode;


/**
 * Declaration object for parameters.
 */
class CParam: public CDecl
{
private:
	CNode*	expression;  ///< expression assigned to parameter.
	int	specify;     ///< non-zero if this is a specparam
	Decl_t	subType;     ///< parameter subtype.
        int     inlined;     ///< non-zero if parameter declaration is
			     ///< inlined in module definition.
public:
	/**
 	 * Create parameter declaration.
 	 * \param symbol name of declaration.
 	 * \param aLoc file coordinates.
 	 * \param type declaration type: eLOCALPARAM, ePARAM.
 	 */
	CParam( CSymbol* symbol, Coord_t* aLoc, Decl_t type );
	/**
 	 * Create a copy of parameter declaration.
 	 * \param param declaration to copy.
 	 */
	CParam( const CParam& param );
	/**
 	 * Determine if width of declaration is constant, ie dependent
 	 * upon only constants and parameters.
 	 * \return non-zero of width of declaration is constant.
 	 */
	virtual int  IsWidthConstant( void );
	/**
 	 * Determine if width of declaration is volatile, ie depend
 	 * upon parameters or variables.
 	 * \return non-zero of width of declaration is volatile.
 	 */
	virtual int  IsWidthVolatile( void );
	/**
 	 * Determine if width of declaration can be evaluated.
 	 * \return non-zero of width of declaration can be evaluated.
 	 */
	virtual int  IsWidthEvaluateable( void );
	/**
 	 * Evaluate width of declaration.
 	 * \return width of declaration.
 	 */
	virtual INT32 GetWidth( void );
	/**
 	 * Get expression representing width of declaration.
 	 * \return expression representing width of declaration.
	 */
	virtual CNode* GetWidthExp( void );
	/**
 	 * Evaluate current decl width direction. Expression must be constant.  
	 * \return 1 - big endian, -1 - little endian, 0 - cannot be determined
	 */
	virtual int   WidthDirection( void );
	/**
 	 * Set specify attribute.
 	 * \param v non-zero if declaration is a specparam.
 	 */
	void 	Specify( int v ) { specify = v; }
	/**
 	 * Get specify attribute.
 	 * \return non-zero if declaration is a specparam.
 	 */
	int	Specify( void ) { return specify; }
	/**
 	 * Set expression assigned to parameter.
 	 * \param aExp value expression.
 	 */
	void 	SetExpression( CNode* aExp );
	/**
 	 * Get expression assigned to parameter.
 	 * \return expression assigned to parameter.
	 */
	CNode*	GetExpression( void );
	/**
 	 * Get node expression type.
 	 * \return expression type.
 	 */
	virtual NodeType_t GetNodeType( void );
	/**
 	 * Dump parameter declaration info to file descriptor.
 	 * \param f file descriptor.
 	 */
	virtual void Dump( FILE* f );
	/**
 	 * Set parameter declaration subtype.
 	 * \param t declaration subtype.
 	 */
	void	SetSubType( Decl_t t ) { subType = t; }
	/**
 	 * Get parameter declaration subtype.
 	 * \return declaration subtype.
 	 */
	Decl_t	GetSubType() { return subType; }
	/**
 	 * Set inlined attribute.
 	 * \param v non-zero if parameter declaration was inlined
 	 *  	    in module definition.
 	 */
        void	SetInlined( int v ) { inlined = v; }
	/**
 	 * Get inlined attribute.
 	 * \return v non-zero if parameter declaration was inlined
 	 *  	    in module definition.
 	 */
        int	GetInlined() { return inlined; }
	/**
 	 * These should not be directly called.
 	 * \cond
 	 */
	void    PreVisit1( int (*func)(CNode*,void*), void* data );
	void    PostVisit1( void (*func)(CNode*, void*), void* data );
	void    PostSubVisit1( CNode* (*func)(CNode*, void*), void* data );
	/**
 	 * \endcond
 	 */
};

#endif // CPARAM_HPP
