/*****************************************************************************
 * Copyright (C) 1997-2007, Mark Hummel
 * This file is part of Vrq.
 *
 * Vrq is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * Vrq is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, 
 * Boston, MA  02110-1301  USA
 *****************************************************************************
 */
/******************************************************************************
 *
 *
 *	   csymtab.hpp
 *		- class definition for symbol tables 
 *
 *
 ******************************************************************************
 */

#ifndef CSYMTAB_HPP
#define CSYMTAB_HPP

#include "glue.h"
#include "cdecl.h"
#include <map>
#include <algorithm>



class CSymbol;
			
/**
 * Aux class used to create symbol table scoping.
 * \cond
 */

template<class T1>
class CSymtabEntry : public map<CSymbol*,T1*>
{
private:
	INT32		depth;
	CSymtabEntry*	previous;
public:

/*********************************************************
	Constructor
**********************************************************/
CSymtabEntry( INT32 aDepth, CSymtabEntry* parent ) 
{
	depth 	     = aDepth;
	previous     = parent;
}
/*********************************************************
	GetDepth
	- return current symbol table depth 
**********************************************************/
INT32	GetDepth() {
	return depth;
}
/*********************************************************
	GetPrevious
	- return previous level 
**********************************************************/
CSymtabEntry*	GetPrevious() {
	return previous;
}

/*********************************************************
	Lookup
	- find symbol by recursively searching table 
**********************************************************/

T1* Lookup( CSymbol* key )
{
	T1*			    result;
	typename map<CSymbol*,T1*>::iterator ptr;

	result = NULL;
	ptr = this->find( key );
	if( ptr != this->end() ) {
		return ptr->second;
	}
	if( previous ) {
		result = previous->Lookup( key );
	}
	return result;
}	

/*********************************************************
	LookupTop
	- find symbol at top level only 
**********************************************************/

T1* LookupTop( CSymbol* key )
{
	T1*	                    result;
	typename map<CSymbol*,T1*>::iterator ptr;

	result = NULL;
	ptr = this->find( key );
	if( ptr != this->end() ) {
		result = ptr->second;
	} 
	return result;
}	

void	Dump( FILE *f, int recurse ) 
{
	typename map<CSymbol*,T1*>::iterator ptr;

	for( ptr = this->begin(); ptr != this->end(); ++ptr) {
		printf( "\t%s => ", ptr->first->GetName() );
		ptr->second->DumpDeclInfo( f );
	}
	if( recurse && previous != NULL ) {
		previous->Dump( f, recurse );
	}
}
		
};
/**
 * \endcond
 */

/**
 * Class for a type safe hierarchical symbol table. Each
 * level of the table represents a different nested scope.
 * Searches a the leaves of the table seach upward to the
 * table root if needed.
 */
template<class T1>
class CSymtab
{
private:
	INT32			currentDepth;	///< current nesting depth
	CSymtabEntry<T1>*	table;		///< current table level
public:

/**
 * Create a symbol table.
 */
CSymtab() 
{
	currentDepth = 0;
	table        = new CSymtabEntry<T1>( 0, NULL );
}
/**
 * Jump back to parent scope.
 */
void PopScope()
{
	MASSERT( currentDepth > 0 );
	MASSERT( table != NULL );

	currentDepth--;
	if( table->GetDepth() <= currentDepth ) {
		return;
	}
	
	table = table->GetPrevious();
}	
/**
 * Create a new scope for table.
 * \param lazy non-zero indicates a new level should only 
 * 	       be created if needed.
 */
void PushScope( int lazy )
{
	currentDepth++;
	if( !lazy ) {
		table = new CSymtabEntry<T1>( currentDepth, table );
	}
}
/**
 * Get current symbol table nesting depth.
 * \return nesting depth (0 is the root).
 */
INT32	GetDepth() { return currentDepth; }

/**
 * Add a symbol and it's assocated object at the current level.
 * Create a new level if needed.
 * \param sym symbol to add to table.
 * \param obj object to associated with symbol.
 */
void	Add( CSymbol* sym, T1* obj )
{
	if( table->GetDepth() < currentDepth ) {
		table = new CSymtabEntry<T1>( currentDepth, table );
	}

	(*table)[sym] = obj;
}
		
/**
 * Lookup symbol only in current scope.
 * \param sym symbol to search for.
 * \return object associated with symbol. NULL if symbol
 * 		  is not found.
 */
T1*	LookupTop( CSymbol* sym )
{
	if( currentDepth != table->GetDepth() ) {
		return NULL;
	}
	return table->LookupTop( sym );
}
		
/**
 * Lookup symbol in all scopes starting at the current scope.
 * \param sym symbol to search for.
 * \return object associated with symbol. NULL if symbol
 * 		  is not found.
 */
T1*	Lookup( CSymbol* sym )
{
	return table->Lookup( sym );
}
		
/**
 * Dump all symbols in table to file descriptor.
 * \param f file descriptor.
 * \param recurse zero if only current depth should be dumped.
 */
void	Dump( FILE *f, int recurse ) 
{
	table->Dump( f, recurse );
}

};

#endif // CSYMTAB_HPP
